/*
 * Copyright (C)2007-2008, IBM, Inc. All rights reserved.
 * IBM, Inc. owns the copyright and other intellectual
 * property rights in this software. Duplication or use of the
 * Software is not permitted except as expressly provided in a
 * written agreement between your company and IBM, Inc.
 */

static int is_same_addr(
		const struct sockaddr *sa1,
		const struct sockaddr *sa2)
{
	if (NULL == sa1 || NULL == sa2 || sa1->sa_family != sa2->sa_family)
		return 0;

	switch (sa1->sa_family) {
		case AF_INET: {
			return 0 == memcmp(
					&((const struct sockaddr_in *)sa1)->sin_addr,
					&((const struct sockaddr_in *)sa2)->sin_addr,
					sizeof(struct in_addr));
		}

#ifdef AF_INET6
		case AF_INET6: {
			return 0 == memcmp(
					&((const struct sockaddr_in6 *)sa1)->sin6_addr,
					&((const struct sockaddr_in6 *)sa2)->sin6_addr,
					sizeof(struct in6_addr));
		}
#endif /* AF_INET6 */

#if defined(AF_LINK) && defined(HAVE_SOCKADDR_DL)
        case AF_LINK: {
			const struct sockaddr_dl *a1 = (const struct sockaddr_dl *)sa1;
			const struct sockaddr_dl *a2 = (const struct sockaddr_dl *)sa2;
			if (a1->sdl_nlen != a2->sdl_nlen || a1->sdl_alen != a2->sdl_alen)
				return 0;
			return 0 == memcmp(
				a1->sdl_data,
				a2->sdl_data,
				a1->sdl_nlen + a1->sdl_alen);
		}
#endif /* AF_LINK && HAVE_SOCKADDR_DL */
	}
	return 0;
}


#ifndef HAVE_INET_NTOP
static const char *inet_ntop(int af, const void *src, char *dst, socklen_t cnt)
{
	switch (af) {
		case AF_INET: {
			struct sockaddr_in in;
			memset(&in, 0, sizeof(in));
			in.sin_family = AF_INET;
			memcpy(&in.sin_addr, src, sizeof(struct in_addr));
			getnameinfo(
					(struct sockaddr *)&in,
					sizeof(struct sockaddr_in),
					dst,
					cnt,
					NULL,
					0,
					NI_NUMERICHOST);
			return dst;
		}
		case AF_INET6: {
			struct sockaddr_in6 in;
			memset(&in, 0, sizeof(in));
			in.sin6_family = AF_INET6;
			memcpy(&in.sin6_addr, src, sizeof(struct in_addr6));
			getnameinfo(
					(struct sockaddr *)&in,
					sizeof(struct sockaddr_in6),
					dst,
					cnt,
					NULL,
					0,
					NI_NUMERICHOST);
			return dst;
		}
	}
	return NULL;
}

static int inet_pton(int af, const char *src, void *dst)
{
	struct addrinfo hints, *res, *res_head;

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = af;

	if (0 != getaddrinfo(src, NULL, &hints, &res_head))
		return -1;

	for (res = res_head; NULL != res; res = res->ai_next)
		memcpy(dst, res->ai_addr, res->ai_addrlen);

	freeaddrinfo(res_head);
	return 0;
}
#endif /* HAVE_INET_NTOP */


static void decode_hwaddr(
		char *buf,
		size_t buflen,
		const unsigned char *hwaddr,
		int hwaddrlen)
{
	char *p = buf;
	size_t limit = buflen / 3;
	int i;

	if (NULL == buf || buflen <= 0)
		return;
	if (hwaddrlen > limit)
		hwaddrlen = limit;
	if (0 == hwaddrlen) {
		*buf = '\0';
		return;
	}
	p += sprintf(p, "%02X", *hwaddr++);
	for (i=1; i<hwaddrlen; ++i)
		p += sprintf(p, ":%02X", *hwaddr++);
}


static void get_net_addr(
		platform_sysconf_net_t *net,
		struct sockaddr *saddr)
{
	switch (saddr->sa_family) {
		case AF_INET:
			if ('\0' == *net->net_ipv4) {
				struct sockaddr_in *saddr4 = (struct sockaddr_in *)saddr;
				inet_ntop(AF_INET,
						&saddr4->sin_addr,
						net->net_ipv4,
						INET_ADDRSTRLEN);
			}
			break;

#ifdef AF_INET6
		case AF_INET6:
			if ('\0' == *net->net_ipv6) {
				struct sockaddr_in6 *saddr6 = (struct sockaddr_in6 *)saddr;
				inet_ntop(AF_INET6,
						&saddr6->sin6_addr,
						net->net_ipv6,
						INET6_ADDRSTRLEN);
			}
			break;
#endif /* AF_INET6 */

#if defined(AF_LINK) && defined(HAVE_SOCKADDR_DL)
		case AF_LINK:
			if ('\0' == *net->net_hwaddr) {
				struct sockaddr_dl *sdl = (struct sockaddr_dl *)saddr;
				const unsigned char *hwaddr =
					((const unsigned char *)sdl->sdl_data) + sdl->sdl_nlen;
				decode_hwaddr(
						net->net_hwaddr,
						BF_HW_ADDRSTRLEN,
						hwaddr,
						sdl->sdl_alen);
			}
			break;
#endif /* AF_LINK && HAVE_SOCKADDR_DL */
	}
}



/*======================================================================
 * WINDOWS
 *======================================================================*/
#ifdef WIN32

/* Win2K does not support GetBestInterfaceEx.  This code ensures that
 * we link to it dynamically.  Remove it when we drop support for Win2K.
 */
#ifdef SUPPORT_WIN2K
typedef DWORD (WINAPI *gbiex_fn)(struct sockaddr *, DWORD *);
static DWORD bf_GetBestInterfaceEx(struct sockaddr *saddr, DWORD *idx)
{
	static gbiex_fn gbiex = NULL;
	if (NULL == gbiex) {
		HMODULE iphlp = GetModuleHandleW(L"iphlpapi.dll");
		if (NULL == iphlp)
			return GetLastError();
		gbiex = (gbiex_fn)GetProcAddress(iphlp, "GetBestInterfaceEx");
		if (NULL == gbiex)
			return GetLastError();
	}
	return (*gbiex)(saddr, idx);
}
#undef GetBestInterfaceEx
#define GetBestInterfaceEx(saddr,idx) \
	bf_GetBestInterfaceEx(saddr,idx)
#endif /* SUPPORT_WIN2K */



/* Win2K does not support GetAdaptersAddresses.  This code ensures that
 * we link to it dynamically.  Remove it when we drop support for Win2K.
 */
#ifdef SUPPORT_WIN2K
typedef ULONG (WINAPI *gaa_fn)(ULONG, ULONG, PVOID, PIP_ADAPTER_ADDRESSES, PULONG);
static ULONG bf_GetAdaptersAddresses(
		ULONG family,
		ULONG flags,
		PVOID rsvd,
		PIP_ADAPTER_ADDRESSES addrs,
		PULONG psz)
{
	static gaa_fn gaa = NULL;
	if (NULL == gaa) {
		HMODULE iphlp = GetModuleHandleW(L"iphlpapi.dll");
		if (NULL == iphlp)
			return GetLastError();
		gaa = (gaa_fn)GetProcAddress(iphlp, "GetAdaptersAddresses");
		if (NULL == gaa)
			return GetLastError();
	}
	return (*gaa)(family, flags, rsvd, addrs, psz);
}
#undef GetAdaptersAddresses
#define GetAdaptersAddresses(family,flags,rsvd,addrs,psz) \
	bf_GetAdaptersAddresses(family,flags,rsvd,addrs,psz)
#endif /* SUPPORT_WIN2K */


/* Look for alternate address.  If we know our IPv4, look for IPv6.
 * If we know our IPv6, look for IPv4...
 */
static void get_net_details_gaa(
		platform_sysconf_net_t *net,
		DWORD idx)
{
	PIP_ADAPTER_ADDRESSES ifaces = NULL;
	PIP_ADAPTER_ADDRESSES iface;
	PIP_ADAPTER_UNICAST_ADDRESS addr;
	ULONG len = 0;
	DWORD ret;
	int i;

	for (i=0; i<8; ++i) {
		ret = (DWORD)GetAdaptersAddresses(AF_UNSPEC, 0, NULL, ifaces, &len);
		if (ERROR_BUFFER_OVERFLOW != ret)
			break;
		if (NULL != ifaces)
			free(ifaces);
		ifaces = malloc(len);
	}
	if (NO_ERROR != ret)
		goto done;

	for (iface = ifaces; NULL != iface; iface = iface->Next) {
		if (idx == iface->IfIndex) {
			for (addr = iface->FirstUnicastAddress;
					NULL != addr;
					addr = addr->Next)
			{
				get_net_addr(net, addr->Address.lpSockaddr);
			}
			break;
		}
	}

done:
	if (NULL != ifaces)
		free(ifaces);
}


static void get_net_details_gbi(
		platform_sysconf_net_t *net,
		const struct sockaddr *saddr,
		socklen_t sa_len)
{
	struct sockaddr saddr_peer;
	int saddr_peer_len = sizeof(saddr_peer);
	MIB_IFROW ifrow;
	ULONG len = 0;
	ULONG ret = 0;

	/* Find the address of our peer. */
	if (0 != getpeername(hRemote, &saddr_peer, &saddr_peer_len))
		return;

	/* Find the matching interface */
	ZeroMemory(&ifrow, sizeof(ifrow));

	ret = GetBestInterfaceEx(&saddr_peer, &ifrow.dwIndex);
	if (NO_ERROR != ret) {
		if (AF_INET == saddr->sa_family) {
			struct sockaddr_in *peer = (struct sockaddr_in *) &saddr_peer;
			ret = GetBestInterface(peer->sin_addr.S_un.S_addr, &ifrow.dwIndex);
		}
		if (NO_ERROR != ret)
			return;
	}

	/* Get the interface's details */
	if (NO_ERROR != GetIfEntry(&ifrow)) {
		if (NULL == net->net_iface) {
			net->net_iface = malloc(16);
			sprintf(net->net_iface, "%lu", ifrow.dwIndex);
		}
		goto no_if_entry;
	}

	if (NULL == net->net_iface) {
		int wlen = wcslen(ifrow.wszName) + 1;
		if (wlen > 1) {
			net->net_iface = malloc(wlen << 1);
			convert_wide_to_utf8(
				ifrow.wszName, wlen,
				net->net_iface, wlen << 1);
		} else {
			wlen = (int)ifrow.dwDescrLen;
			if (wlen > 0) {
				net->net_iface = malloc(wlen << 2);
				wlen = convert_native_to_utf8(
					(const char *)ifrow.bDescr, wlen,
					net->net_iface, wlen << 2);
				net->net_iface[wlen] = '\0';
			}
		}
	}

	if (NULL == net->net_hwaddr) {
		/* FIXME - we should probably check dwType instead of or
		 * in addition to dwPhysAddrLen ...
		 */
		if (ifrow.dwPhysAddrLen <= 6) {
			decode_hwaddr(
					net->net_hwaddr,
					BF_HW_ADDRSTRLEN,
					(const unsigned char *)ifrow.bPhysAddr,
					(int)ifrow.dwPhysAddrLen);
		}
	}

	net->net_speed = ifrow.dwSpeed / 1000000;

no_if_entry:
	get_net_details_gaa(net, ifrow.dwIndex);
}
#endif /* WIN32 */

/*======================================================================*/

/*======================================*/
/* Get network info by using getifaddrs */
/*======================================*/

#ifdef HAVE_GETIFADDRS
static void get_net_details_ifaddrs(
		platform_sysconf_net_t *net,
		const struct sockaddr *saddr,
		socklen_t sa_len)
{
	struct ifaddrs *ifa_head, *ifa;

	/* Look for alternate address.  If we know our IPv4, look for IPv6.
	 * If we know our IPv6, look for IPv4...
	 */
	if (0 != getifaddrs(&ifa_head))
		return;

	/* First pass: find the matching interface name */
	if (NULL == net->net_iface) {
		for (ifa = ifa_head; NULL != ifa; ifa = ifa->ifa_next) {
			if (is_same_addr(saddr, ifa->ifa_addr)) {
				net->net_iface = strdup(ifa->ifa_name);
				break;
			}
		}
	}

#if defined(AF_INET6) || (defined(AF_LINK) && defined(HAVE_SOCKADDR_DL))
	/* Second pass: find other addresses */
	if (NULL != net->net_iface) {
		for (ifa = ifa_head; NULL != ifa; ifa = ifa->ifa_next) {
			if (0 == strcmp(ifa->ifa_name, net->net_iface))
				get_net_addr(net, ifa->ifa_addr);
		}
	}
#endif /* AF_INET6 || (AF_LINK && HAVE_SOCKADDR_DL) */

	freeifaddrs(ifa_head);
}


/*===========================================*/
/* Get network info by using SIOCGIFCONF     */
/* (getifaddrs is preferred, if we have it.) */
/*===========================================*/
#elif defined(SIOCGIFCONF)
static struct ifconf *get_ifconf(void) {
	struct ifconf *ifc = calloc(1, sizeof(struct ifconf));
	size_t len;

#ifdef SIOCGSIZIFCONF
	len = 0;
	if (ioctl(hRemote, SIOCGSIZIFCONF, &len) >= 0 && len > 0) {
		++len;
	} else {
		len = (sizeof(struct ifreq) + 1) << 3;
	}
#else
	len = (sizeof(struct ifreq) + 1) << 3;
#endif /* SIOCGSIZIFCONF */

	if (NULL == ( ifc->ifc_req = malloc(len) )) {
		free(ifc);
		return NULL;
	}

	while (1) {
		ifc->ifc_len = len;
		if (ioctl(hRemote, SIOCGIFCONF, ifc) < 0) {
			free(ifc->ifc_req);
			free(ifc);
			return NULL;
		}
		if (ifc->ifc_len < len)
			break;
		len <<= 1;
		ifc->ifc_req = realloc(ifc->ifc_req, len);
		if (NULL == ifc->ifc_req) {
			free(ifc);
			return NULL;
		}
	}

	if (ifc->ifc_len > 0) {
		ifc->ifc_req = realloc(ifc->ifc_req, ifc->ifc_len);
		if (NULL == ifc->ifc_req) {
			free(ifc);
			return NULL;
		}
	}

	return ifc;
}



static void get_net_details_ifconf(
		platform_sysconf_net_t *net,
		const struct sockaddr *saddr,
		socklen_t sa_len)
{
	const char *cp, *cp_end;
	struct ifconf *ifc;
	struct ifreq *ifr;

	if (NULL == ( ifc = get_ifconf() ))
		return;

	/* First pass: find the matching interface name */
	if (NULL == net->net_iface) {
		cp = (const void *)ifc->ifc_req;
		cp_end = &cp[ifc->ifc_len];
		while (cp < cp_end) {
			ifr = (struct ifreq *)cp;
			if (is_same_addr(saddr, &ifr->ifr_addr)) {
				net->net_iface = strdup(ifr->ifr_name);
				break;
			}
			cp += sizeof(ifr->ifr_name) + BF_SA_LEN(&ifr->ifr_addr);
		}
		cp = (const void *)ifc->ifc_req;
	}

#if defined(AF_INET6) || defined(AF_LINK)
	/* Second pass: find other addresses */
	if (NULL != net->net_iface) {
		cp = (const void *)ifc->ifc_req;
		cp_end = &cp[ifc->ifc_len];
		while (cp < cp_end) {
			ifr = (struct ifreq *)cp;
			get_net_addr(net, &ifr->ifr_addr);
			cp += sizeof(ifr->ifr_name) + BF_SA_LEN(&ifr->ifr_addr);
		}
	}
#endif /* AF_INET6 || AF_LINK */

	free(ifc->ifc_req);
	free(ifc);
	return;
}

#endif /* HAVE_GETIFADDRS / SIOCGIFCONF / misc ioctls */


#ifdef SIOCGIFNAME
/* These are last resorts to try if HAVE_GETIFADDRS and SIOCGIFCONF are
 * unavailable (or in the case of the MAC address, if AF_LINK is not
 * used by this platform).
 */
static void get_net_details_ifname(
	platform_sysconf_net_t *net,
	const struct sockaddr *saddr,
	socklen_t sa_len)
{
	struct ifreq ifr;
	if (NULL != net->net_iface)
		return;
	memcpy(&ifr.ifr_addr, saddr, sa_len);
	if (ioctl(hRemote, SIOCGIFNAME, &ifr) >= 0) {
		net->net_iface = strdup(ifr.ifr_name);
		return;
	}
}
#endif /* SIOCGIFNAME */


#ifdef SIOCGIFHWADDR
static void get_net_details_hwaddr(
	platform_sysconf_net_t *net,
	const struct sockaddr *saddr,
	socklen_t sa_len)
{
	struct ifreq ifr;
	if (NULL == net->net_iface || '\0' != *net->net_hwaddr)
		return;
	strcpy(ifr.ifr_name, net->net_iface);
	if (ioctl(hRemote, SIOCGIFHWADDR, &ifr) >= 0) {
		decode_hwaddr(
				net->net_hwaddr, 
				BF_HW_ADDRSTRLEN,
				(const unsigned char *)ifr.ifr_addr.sa_data,
				6);
		return;
	}
}
#endif /* SIOCGIFHWADDR */


platform_sysconf_net_t *platform_sysconf_net_new(void) {
	platform_sysconf_net_t *net;
	struct sockaddr sa;
	socklen_t sa_len = sizeof(sa);

	DebugBlock("CRF: platform_sysconf_net_new");
	if (!platform_protocol_has_socket())
		return NULL;
	if (getsockname(hRemote, &sa, &sa_len) < 0)
		return NULL;
	if (NULL == ( net = calloc(1, sizeof(*net)) ))
		return NULL;

	switch (sa.sa_family) {
		case AF_INET: {
			struct sockaddr_in *saddr4 = (struct sockaddr_in *) &sa;
			inet_ntop(AF_INET, &saddr4->sin_addr,
					net->net_ipv4, INET_ADDRSTRLEN);
			net->net_ipv = 4;
			break;
		}

#ifdef AF_INET6
		case AF_INET6: {
			struct sockaddr_in6 *saddr6 = (struct sockaddr_in6 *) &sa;
			inet_ntop(AF_INET6, &saddr6->sin6_addr,
					net->net_ipv6, INET6_ADDRSTRLEN);
			net->net_ipv = 6;
			break;
		}
#endif /* AF_INET6 */

		default:
			free(net);
			return NULL;
	}

	/* Get FQDN */
	net->net_fqdn = malloc(NI_MAXHOST);
	if (NULL != net->net_fqdn) {
		if (0 != getnameinfo(&sa, sa_len,
				net->net_fqdn, NI_MAXHOST,
				NULL, 0, 0))
		{
			free(net->net_fqdn);
			net->net_fqdn = NULL;
		}
	}

	/* Get interface data */
#ifdef WIN32
	if (platform_level_xp())
		get_net_details_gbi(net, &sa, sa_len);
#endif /* WIN32 */
#ifdef HAVE_GETIFADDRS
	get_net_details_ifaddrs(net, &sa, sa_len);
#elif defined(SIOCGIFCONF)
	get_net_details_ifconf(net, &sa, sa_len);
#endif /* HAVE_GETIFADDRS / SIOCGIFCONF / misc ioctls */
#ifdef SIOCGIFNAME
	get_net_details_ifname(net, &sa, sa_len);
#endif /* SIOCGIFNAME */
#ifdef SIOCGIFHWADDR
	get_net_details_hwaddr(net, &sa, sa_len);
#endif /* SIOCGIFHWADDR */
	return net;
}
